# Keylime logging configuration

# The configuration file version number
[logging]
version = {{ logging.version }}

[loggers]
keys = {{ logging.loggers_keys }}

[handlers]
keys = {{ logging.handlers_keys }}

[formatters]
keys = {{ logging.formatters_keys }}

[formatter_formatter]
format = {{ logging.formatter_formatter_format }}
datefmt = {{ logging.formatter_formatter_datefmt }}

[logger_root]
level = {{ logging.logger_root_level }}
handlers = {{ logging.logger_root_handlers }}

[handler_consoleHandler]
class = {{ logging.handler_consoleHandler_class }}
level = {{ logging.handler_consoleHandler_level }}
formatter = {{ logging.handler_consoleHandler_formatter }}
args = {{ logging.handler_consoleHandler_args }}

[logger_keylime]
level = {{ logging.logger_keylime_level }}
qualname = {{ logging.logger_keylime_qualname }}
handlers = {{ logging.logger_keylime_handlers }}
